@@ -26,7 +26,10 @@ pub struct Spamc {
 }
 
 impl Spamc {
-    const SPAMC_PROGRAM: &'static str = "spamc";
+    const SPAMC_PROGRAM: &'static str = match option_env!("SPAMASSASSIN_MILTER_SPAMC") {
+        Some(p) => p,
+        None => "/usr/bin/spamc",
+    };
 
     pub fn new(spamc_args: &'static [String]) -> Self {
         Self {
@@ -41,7 +44,7 @@ impl Process for Spamc {
     fn connect(&mut self) -> Result<()> {
         // `Command::spawn` always succeeds when `spamc` can be invoked, even if
         // logically the command is invalid, eg if it uses non-existing options.
-        let mut spamc = Command::new(Spamc::SPAMC_PROGRAM)
+        let mut spamc = Command::new(Self::SPAMC_PROGRAM)
             .args(self.spamc_args)
             .stdin(Stdio::piped())
             .stdout(Stdio::piped())
@@ -172,29 +175,35 @@ impl Client {
         Ok(())
     }
 
-    pub fn send_forged_received_header(
+    pub fn send_synthesized_received_header(
         &mut self,
         helo_host: &str,
         client_name_addr: &str,
         my_hostname: &str,
         mta: &str,
         tls: bool,
+        auth: bool,
         queue_id: &str,
         date_time: &str,
     ) -> Result<()> {
+        // Sending this ‘Received’ header is crucial: Milters don’t see the
+        // MTA’s own ‘Received’ header. However, SpamAssassin draws a lot of
+        // information from that header. So we make one up and send it along.
+
         let buf = format!(
-            "Received: from {} ({})\r\n\
-             \tby {} ({}) with {} id {};\r\n\
-             \t{}\r\n\
-             \t(envelope-from {})\r\n",
-            helo_host,
-            client_name_addr,
-            my_hostname,
-            mta,
-            if tls { "ESMTPS" } else { "ESMTP" },
-            queue_id,
-            date_time,
-            self.sender
+            "Received: from {helo} ({client})\r\n\
+             \tby {hostname} ({mta}) with ESMTP{tls}{auth} id {id};\r\n\
+             \t{date_time}\r\n\
+             \t(envelope-from {sender})\r\n",
+            helo = helo_host,
+            client = client_name_addr,
+            hostname = my_hostname,
+            mta = mta,
+            tls = if tls { "S" } else { "" },
+            auth = if auth { "A" } else { "" },
+            id = queue_id,
+            date_time = date_time,
+            sender = self.sender
         );
 
         self.process.writer().write_all(buf.as_bytes())?;
@@ -333,7 +342,7 @@ fn replace_body(
 #[cfg(test)]
 mod tests {
     use super::*;
-    use std::{cell::RefCell, collections::HashSet};
+    use std::cell::RefCell;
 
     #[derive(Clone, Debug, Default, Eq, Hash, Ord, PartialEq, PartialOrd)]
     struct MockSpamc {
@@ -380,6 +389,7 @@ mod tests {
     #[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
     enum Action {
         AddHeader(String, String),
+        InsertHeader(usize, String, String),
         ReplaceHeader(String, usize, Option<String>),
         AppendBodyChunk(Vec<u8>),
         SetErrorReply(String, Option<String>, Vec<String>),
@@ -403,6 +413,12 @@ mod tests {
             ))
         }
 
+        fn insert_header(&self, index: usize, name: &str, value: &str) -> milter::Result<()> {
+            Ok(self.called.borrow_mut().push(
+                Action::InsertHeader(index, name.to_owned(), value.to_owned())
+            ))
+        }
+
         fn replace_header(&self, name: &str, index: usize, value: Option<&str>) -> milter::Result<()> {
             Ok(self.called.borrow_mut().push(
                 Action::ReplaceHeader(name.to_owned(), index, value.map(|v| v.to_owned()))
@@ -427,10 +443,6 @@ mod tests {
             unimplemented!();
         }
 
-        fn insert_header(&self, _: usize, _: &str, _: &str) -> milter::Result<()> {
-            unimplemented!();
-        }
-
         fn quarantine(&self, _: &str) -> milter::Result<()> {
             unimplemented!();
         }
@@ -468,7 +480,7 @@ mod tests {
         assert_eq!(client.bytes_written(), 48);
         assert_eq!(
             as_mock_spamc(client.process.as_ref()).buf,
-            Vec::from(b"name1: value1\r\nname2: value2\r\n\tcontinued\r\n\r\nbody" as &[_])
+            b"name1: value1\r\nname2: value2\r\n\tcontinued\r\n\r\nbody".as_ref()
         );
     }
 
@@ -489,12 +501,12 @@ mod tests {
 
         assert_eq!(
             as_mock_spamc(client.process.as_ref()).buf,
-            Vec::from(format!(
+            format!(
                 "X-Envelope-From: {}\r\n\
                  X-Envelope-To: {},\r\n\
                  \t{}\r\n",
                 sender, recipient1, recipient2
-            ).as_bytes())
+            ).as_bytes()
         );
     }
 
@@ -522,16 +534,15 @@ mod tests {
         let status = client.process("id", &actions, &config).unwrap();
 
         assert_eq!(status, Status::Reject);
-
-        let called = actions.called.borrow();
-        assert_eq!(called.len(), 1);
         assert_eq!(
-            called.first().unwrap(),
-            &Action::SetErrorReply(
-                "550".into(),
-                Some("5.7.1".into()),
-                vec!["Spam message refused".into()]
-            )
+            actions.called.borrow().as_slice(),
+            [
+                Action::SetErrorReply(
+                    "550".into(),
+                    Some("5.7.1".into()),
+                    vec!["Spam message refused".into()],
+                ),
+            ]
         );
     }
 
@@ -550,19 +561,16 @@ mod tests {
         let status = client.process("id", &actions, &config).unwrap();
 
         assert_eq!(status, Status::Accept);
-
-        let called = actions.called.borrow();
-        assert_eq!(called.len(), 4);
-
-        let mut expected = HashSet::new();
-        expected.insert(Action::AddHeader("X-Spam-Flag".into(), " YES".into()));
-        expected.insert(Action::ReplaceHeader("X-Spam-Level".into(), 1, Some(" *****".into())));
-        expected.insert(Action::ReplaceHeader("X-Spam-Report".into(), 1, None));
-        expected.insert(Action::AppendBodyChunk(b"Report".to_vec()));
-
-        for a in called.iter() {
-            assert!(expected.contains(&a));
-        }
+        assert_eq!(
+            actions.called.borrow().as_slice(),
+            [
+                Action::ReplaceHeader("X-Spam-Level".into(), 1, None),
+                Action::InsertHeader(0, "X-Spam-Level".into(), " *****".into()),
+                Action::InsertHeader(0, "X-Spam-Flag".into(), " YES".into()),
+                Action::ReplaceHeader("x-spam-report".into(), 1, None),
+                Action::AppendBodyChunk(b"Report".to_vec()),
+            ]
+        );
     }
 
     #[test]
@@ -581,7 +589,7 @@ mod tests {
         assert_eq!(status, Status::Accept);
 
         let called = actions.called.borrow();
-        assert!(called.contains(&Action::AddHeader("X-Spam-Level".into(), " *****".into())));
+        assert!(called.contains(&Action::InsertHeader(0, "X-Spam-Level".into(), " *****".into())));
         assert!(!called.contains(&Action::AppendBodyChunk(b"Report".to_vec())));
     }
 }
