@@ -10,20 +10,18 @@ fn ham_message() {
     let config = builder.build();
 
     let server = spawn_mock_spamd_server(SPAMD_PORT, |ham| {
-        Ok(ham
-            .replacen(
-                "X-Spam-Checker-Version: BogusChecker 1.0.0\r\n",
-                "X-Spam-Checker-Version: MyChecker 1.0.0\r\n",
-                1,
-            )
-            .replacen("X-Spam-Report: Bogus report\r\n", "", 1)
-            .replacen(
-                "\r\n\r\n",
-                "\r\n\
-                 X-Spam-Custom: Custom-Value\r\n\
-                 \r\n",
-                1,
-            ))
+        let mut ham = ham
+            .replacen("X-Spam-Checker-Version: BogusChecker 1.0.0\r\n", "", 1)
+            .replacen("X-Spam-Report: Bogus report\r\n", "", 1);
+
+        // Prepend replaced and newly added SpamAssassin headers.
+        ham.insert_str(
+            0,
+            "X-Spam-Checker-Version: MyChecker 1.0.0\r\n\
+             X-Spam-Custom: Custom-Value\r\n",
+        );
+
+        Ok(ham)
     });
     let miltertest = spawn_miltertest_runner(file!());
 
